/*

Serielles Display mit @ als Steuerzeichen
Schnittstelle: 4800 8N1
  by Dietmar Schroeder 2011 

Kennt folgende Steuersequenzen:
  @c  Clear+Home1
  @h  home1 (Angang 1. Zeile)
  @H  home2 (Angang 2. Zeile)

Sequenzen, die in der Waschmaschine nicht benutzt werden:  
  @l  Licht an
  @0  Licht aus
  @bX PWM-Wert (Helligkeit) X ist ein Kleinbuchstabe: a=0% ... q=100%
  @i  Display neu initialisieren
  @@  Klammeraffe anzeigen
  @2  2ms delay
  @!  nix machen (keep alive Message)
  Wird 10s nix empfangen, wird ein Softwareeset ausgelst

*/

/** \file
Serial interface for LCD panel.
Uses PIC 16c73B as microcontroller

The pin connection is as follows:
\verbatim

     (ai=Analog In, do=digital out  dio=digital in/out  di=digital in)

     A0  (ai)  Light Sensor
     A1  (ai)  -
     A2  (di)  
     A3  (ai)  -
     A4  (di)  
     A5  (di)  

     B0  (do) LCD D0
     B1  (do) LCD D1
     B2  (do) LCD D2
     B3  (do) LCD D3
     B4  (do) LCD D4
     B5  (do) LCD D5
     B6  (do) LCD D6
     B7  (do) LCD D7

     C0  (do)  LCD enable
     C1  (do)  LCD rs
     C2  (do)  LCD Backlight (PWM)
     C3  (di)  
     C4  (do)  
     C5  (di)  
     C6  (do)  XMIT serial
     C7  (di)  RCV serial

\endverbatim
*/


// ----------------------------------------------------------------------------
//  PIC Programmer
// ----------------------------------------------------------------------------
#device PIC16C73B
#include <16C73.H>

// HighSpeed,PowerUpTimer,WatchdogTimer,NoProtection,BrownoutDetection
//#fuses HS,NOPROTECT,WDT,PUT,BROWNOUT <<- Brownout nimmt der Compiler nicht :-(
#fuses HS,PUT,WDT,NOPROTECT


// ----------------------------------------------------------------------------
//  Constants and compiler options
// ----------------------------------------------------------------------------
#define CLOCK_SPEED 4000000  // 4MHz

// need to know the clock speed for UART
#use delay(clock=CLOCK_SPEED)

// we never change the port direction, so we use fast mode I/O.
#use FAST_IO(A)
#use FAST_IO(B)
#use FAST_IO(C)
#priority rda

// T1 at internal clock (=f_Oscillator/4)
#define T1_RESET_VAL (65413)      // overflowValue - CLOCK_SPEED/4/prescale/periodsPerSecond
                                  // = (65536 -    4000000  /4/   8       /1000)         =
                                  // = (65536 - 125)
                                  // = 65411, nehme 2 mehr sonst 5% Fehler


#use rs232(baud=4800, xmit=PIN_C6, rcv=PIN_C7, PARITY=N, BITS=8)


// ----------------------------------------------------------------------------
//  Port maps
// ----------------------------------------------------------------------------

struct portAMap {
    boolean lightSensor;
    boolean unused1;
    boolean keyA;
    boolean unused3;
    boolean keyB;
    boolean keyC;
  }portA;

struct portBMap {
    int data : 8;
  }portB;

struct portCMap {
    boolean lcdEnable;
    boolean lcdRS;
    boolean light;
    boolean debug;
    boolean ResetSunString;
    boolean keyD;
    boolean xmit;
    boolean rcv;
  }portC;

#byte portA = 5                        // This puts the entire structure
                                       // on to port A (at address 5)
#byte portB = 6                        // This puts the entire structure
                                       // on to port B (at address 6)
#byte portC = 7                        // This puts the entire structure
                                       // on to port C (at address 7)

#byte adcResultRegister = 0x1e    //AD-Wandler Ergebnis
#byte adcon0Register    = 0x1f    //AD-Wandler Kontrollregister
#byte rcstaRegister     = 0x18    //UART Status and Control Register
#byte rcregRegister     = 0x1a    //UART Empfangspuffer
#byte pie1Register      = 0x8c    //Interrupt enable

STRUCT portAMap const TRIS_A = {1,1,1,1,1,1};     // All IN
STRUCT portBMap const TRIS_B = {0};               // all OUT
STRUCT portCMap const TRIS_C = {0,0,0,1,0,1,0,1}; // IN on Debug, keyD and rcv


// ----------------------------------------------------------------------------
//  Global variables
// ----------------------------------------------------------------------------
                            // v--- we us all spare memory for RX buffer!
#define RX_BUFFER_SIZE        70  //!< receive fifo  (Controller -> PIC)

struct globVar{
  byte rxBuffer[RX_BUFFER_SIZE]; //!< input fifo

  byte rxNextIn;  //!< fifo pointer
  byte rxNextOut; //!< fifo pointer

  byte rsLine;      //!< 1=write to LCD display 0=Write to LCD register
  byte brightness;  //!< 0=0ff 32=100%. Display light=100% at power up

  long int tic;                  //!< Incremented every ms by Timer1-ISR
}glob;

// ----------------------------------------------------------------------------
//! Prozessor neu starten
void softReset(){
#asm
  goto 0
#endasm
}

// ----------------------------------------------------------------------------
//  FIFO
// ----------------------------------------------------------------------------

// -------------------------------------
//! Check whether RX data available
// -------------------------------------
#define rxDataAvail (glob.rxNextIn != glob.rxNextOut)


// ---------------------------------------------------------------------------
/*! Get byte from serial input fifo.
    Blocks until character available.
*/// -------------------------------------------------------------------------
byte rxGetC() {
   byte c;

   while(!rxDataAvail) ;
   c=glob.rxBuffer[glob.rxNextOut];
   glob.rxNextOut++;
   if (RX_BUFFER_SIZE==glob.rxNextOut){
     glob.rxNextOut = 0;
   }

   return(c);
}

// ----------------------------------------------------------------------------
//   LCD functions
// ----------------------------------------------------------------------------


// ----------------------------------------------------------------------------
/*! Setze globale Helligkeitsvariable.
    Prft Parametergrenzen.
*/
// ----------------------------------------------------------------------------
void setBrightness(byte h){
  if (h<0) h=0;
  if (h>16) h=16;
  glob.brightness = h*8;
}

// ---------------------------------------------------------------------------
/*! Send byte to display.
    Depending on the value of glob.rsLine the byte is written to display(glob.rsLine=1) or register (glob.rsLine=0)

   \param b is the byte to be send.
   \note the BUSY-Flag isn't ckecked (since we can't read from display with R/W hard-wired to GND)
*/// -------------------------------------------------------------------------
void lcdSendByte(byte b ) {

      portC.lcdRS = glob.rsLine;
      delay_cycles(1);  //t_AS: 40ns min  #BATRON, page 57
      portC.lcdEnable = 1;
      //delay_cycles(4);  // not necessary

      portB.data = b;
      delay_cycles(1); //t_DSW: 80ns min.

      portC.lcdEnable = 0;
      delay_us(1); //t_H: 10ns min

      delay_us(1);  //t_CYC 500ns min

      // CAUTION: we do not check the BUSY-Flag (since we can't read from display)
      // CAUTION: so for HOME and CLEAR commands the user has to add 2ms delay!


      //500ns seems too short for t_CYC. Display somtimes didn't catch a char.
      //so we delay a bit more...
      delay_us(100);
}



// ---------------------------------------------------------------------------
/*! switch LCD backlight on.
    Also restart the light timer and send "light on" code on serial interface.
*/// -------------------------------------------------------------------------
void switchOnLight(){

     //swicht on the light
     set_pwm1_duty(glob.brightness);

     // This sets the time the pulse is high each cycle.
     // the high time will be:
     //    value*4*(1/clock)*t2div
     // highTime   = (1/4000000)*16*4  * value

     // cycle time = (1/4000000)*16*128 = 488 Hz
     // so in theory value=32 for 100%  (0=0%)
     // but we found that value=128 for 100%  (0=0%)

}

// ---------------------------------------------------------------------------
/*! switch LCD backlight off (0%).
*/// -------------------------------------------------------------------------
void switchOffLight(){
  //swicht off the light
  set_pwm1_duty(0);
}

// ---------------------------------------------------------------------------
/*! Clear LCD and set cursor home
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------

void clearLcd() {
    glob.rsLine=0;
    lcdSendByte(0x01);
    glob.rsLine=1;
    delay_ms(2);

}

// ---------------------------------------------------------------------------
/*! Set cursor home
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------
void home1Lcd() {
    glob.rsLine=0;
    lcdSendByte(0x02);
    glob.rsLine=1;
    delay_ms(2);
}
// ---------------------------------------------------------------------------
/*! Set cursor at start of Line 2
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------
void home2Lcd() {
    glob.rsLine=0;
    lcdSendByte(192);
    glob.rsLine=1;
    delay_ms(2);
}

// ---------------------------------------------------------------------------
/*! Show SW-Version on LCD (at startup)
*/// -------------------------------------------------------------------------
void lcdWelcomeMessage(){
    clearLcd();
    printf(lcdSendByte,"  Zabex LCD 1.0"); // Hier die aktuelle Versionsnummer eintragen
}

// ---------------------------------------------------------------------------
/*! Initialize display (2 Lines, 8 Bit, no shift, increment)
   (HD 44780 based displays only)
   Writes "LCD Ready" on display after initialization
*/// -------------------------------------------------------------------------
void lcdInit() {
    byte i;

    portC.lcdRs = 0;
    portC.lcdEnable = 0;
    glob.rsLine=0;

    restart_wdt();
    delay_ms(15);
    restart_wdt();

    lcdSendByte(0x30); // DB5=1 DB4=1        # see BATRON-Specification for HD44780, page 58
    delay_ms(5);      // wait more than 4.1ms

    lcdSendByte(0x30); // DB5=1 DB4=1
    delay_ms(1);       // wait more than 100us

    lcdSendByte(0x30);
    delay_ms(2);	   //just to be sure...


    lcdSendByte(0x38); //2 Lines, Font 0 (5x7 dot)  N=1 F=0
    lcdSendByte(0x0c); // Display on
    delay_us(2);

    lcdSendByte(0x01); // Display clear
    delay_ms(3);

    lcdSendByte(0x06); // Entry mode; no shift
    delay_ms(2);

    glob.rsLine=1;
    restart_wdt();
}

// ---------------------------------------------------------------------------
/*! Send value to LCD register
   (HD 44780 based displays only)
   The RS-line is restored after writing.
*/// -------------------------------------------------------------------------
void sendToRegister(byte b){
  byte oldRS;
  oldRS = glob.rsLine;
  glob.rsLine = 0;
  lcdSendByte(b);
  glob.rsLine = oldRS;
  delay_ms(3);
}
// ---------------------------------------------------------------------------
/*! Send char to display or handle special sequence
    if a \t is received the next character is handled specially. If too many non-alphanum characters
    are received, a soft-reset is performed (bad serial communication detected)

    \param c character to be interpreted.
*/// -------------------------------------------------------------------------

void interpreteChar( char c) {
static int sonderzeichen=0;

   if (c <32 || c > 127){
     sonderzeichen++;
     if (sonderzeichen > 9){
       //Hm, so viele Sonderzeichen? Da muss die serielle Schnittstelle was falsch machen.
       //Es wird jetzt ein RESET ausgelst
       softReset();
     }
   }else{
     sonderzeichen = 0;
   }


   if ('@' == c){
     // escape sequence
     c=rxGetC();
     switch (c) {
       case 'l' : switchOnLight();                  break;  // display light ON
       case '0' : switchOffLight();                 break;  // display light OFF immediately
       case 'b' : c=rxGetC(); setBrightness(c - 97);     break;  // PWM-Value for light ON (a=0% ... q=100%)
       case '2' : delay_ms(2);                      break;  // wait 2 ms
       case 'i' : lcdInit();                        break;  // standard init for HD 44780 driven displays
       case 'c' : clearLcd();                       break;
       case 'h' : home1Lcd();                       break;  // 1.Zeichen in 1. Zeile
       case 'H' : home2Lcd();                       break;  // 1.Zeichen in 2. Zeile
       case '!' :                                   break;  // Keep alive Message. (Ruecksetzen timeout)
       case '@': lcdSendByte('@');                break;
     }
   }else{
     lcdSendByte(c);
   }
}


// ----------------------------------------------------------------------------
//  Interrupt service routines
// ----------------------------------------------------------------------------

// -------------------------------------
#int_default
void doNothing(){

}

#int_timer1
void t1_isr(){
  set_timer1(T1_RESET_VAL);
  glob.tic++;
}

// -------------------------------------
/*! Interrupt service function for serial interface.
    Stores incoming character in rx-fifo.
   ------------------------------------- */
#int_rda
void rx_isr() {
//static int c4;

  int newPosition;
  int dummy;

   newPosition = glob.rxNextIn + 1;
   if  (RX_BUFFER_SIZE == newPosition) {
     newPosition = 0;
   }
   if (newPosition!= glob.rxNextOut) {
     glob.rxBuffer[glob.rxNextIn]=getc();
     glob.rxNextIn = newPosition;           // Buffer not full !!
   }else{
     bit_clear(rcstaRegister,1); //Clear OERR  (Overrun Error)
     //read receive register
     dummy = rcregRegister;
   }
}


// ---------------------------------------------------------------------------
/*! Preset all global variables
*/// -------------------------------------------------------------------------
void presetGlobalVar(){
  glob.rxNextIn    =   0;  /*! Positionen in den FIFOs */
  glob.rxNextOut   =   0;
  glob.rsLine      =   1;    // Default: write to display (not register)
  glob.brightness  =   128;  // 0=0ff 128=100%.
}

// ----------------------------------------------------------------------------
//  MAIN
// ----------------------------------------------------------------------------
int main(){
int i;
long timeout;  //Wenn innerhalb von 10s nichts von Serieller kommt, wird Reset ausgelst.

  //#priority rda, tbe, ad

  timeout=0;
  disable_interrupts(GLOBAL);
  presetGlobalVar();

  // Setup ports
  set_tris_A(TRIS_A);
  set_tris_B(TRIS_B);
  set_tris_C(TRIS_C);

  // Setup timer
  setup_timer_1(T1_INTERNAL | T1_DIV_BY_8);
  set_timer1(T1_RESET_VAL);

  // Setup PWM
  setup_ccp1(CCP_PWM);   // Configure CCP1 as a PWM
  set_pwm1_duty(128);

  setup_timer_2(T2_DIV_BY_16, 127, 1);
  // The cycle time will be (1/clock)*4*t2div*(period+1)
  // In this program clock=4000000 and period=127
  //  cycle time = (1/4000000)*16*1*128 =^ 488 Hz


  // Setup interrupts
  //  enable_interrupts(INT_ADC); NO. We need no interrupts
  enable_interrupts(INT_RDA);  //Serial interface
  enable_interrupts(INT_TIMER1);
  disable_interrupts(INT_TIMER2);
  disable_interrupts(INT_RTCC);

  enable_interrupts(GLOBAL);

  lcdInit();

  lcdWelcomeMessage();
  switchOnLight(); // Default at startup

  // ---- Main Loop ----
  while(1){

    //handle data in TX and RX FIFO
    if (rxDataAvail) {
      interpreteChar(rxGetC());
      timeout=0;
    }

    //Check whether a millisecond has passed  (tic is incremented by TIMER1 interrupt)
    while (glob.tic){
      glob.tic--;
      timeout++;
      //RS232 im Fehlerfall wiederbeleben
      if (bit_test(rcstaRegister, 1) || bit_test(rcstaRegister, 2) ){  //OERR oder FERR
        bit_clear(rcstaRegister,1); //Clear OERR
        bit_clear(rcstaRegister,4);
        bit_set(rcstaRegister,4);
      }
      bit_set(pie1Register, 5); //enable UART Interrupts

      //throw a bone to the watchdog to keep it satisfied
      if (timeout < 1000* 10) restart_wdt();
    }
  }

  return 0;
}
// EOF

